#!/bin/sh
#
# Disable ACPI for the next boot. This is similar to using QEMU's '-no-acpi'
# option. Also reconfigure grub to use serial port ttyAMA0 instead of ttyS0.
#

set -eu

if ! [ -d /boot/efi/EFI ] ; then
	echo "No such directory: /boot/efi/EFI" >&2
	exit 1
fi

if [ "${1:-}" = "cleanup" ] && ! [ -e /boot/efi/ForceNoAcpi ] ; then
	echo "No cleanup necessary, exiting..."
	exit 0
fi

apt-get install -y efibootmgr

# This will fail and terminate the script if EFI variables are not supported
efibootmgr >/dev/null || exit 0

# Get the boot number of the EFI Shell
shell_bootnum=$(efibootmgr | grep -E '^Boot[0-9A-F]{4}\*? +EFI Internal Shell$' | head -c 8 | tail -c +5 || true)
if [ -z "${shell_bootnum}" ] ; then
	efibootmgr
	echo "Failed to determine EFI Shell boot number" >&2
	exit 1
fi

if [ "${1:-}" = "cleanup" ] ; then
	# Create startup.nsh which (re-)enables ACPI
	cat > /boot/efi/startup.nsh <<EOF
# Enable ACPI (takes effect after reset)
setvar -nv -bs -guid 50bea1e5-a2c5-46e9-9b3a-59596516b00a ForceNoAcpi =

# Clean up
rm fs0:\ForceNoAcpi

# Remove ourselves to make this a one-shot operation
rm fs0:\startup.nsh

# Reset for the change to take effect
reset
EOF

	# Undo grub config changes
	rm -f /etc/default/grub.d/99-qemu-efi-noacpi.cfg
else
	# Create startup.nsh which disables ACPI
	cat > /boot/efi/startup.nsh <<EOF
# Disable ACPI (takes effect after reset)
setvar -nv -bs -guid 50bea1e5-a2c5-46e9-9b3a-59596516b00a ForceNoAcpi =01

# Indicate that we ran
echo 1 >a fs0:\ForceNoAcpi

# Remove ourselves to make this a one-shot operation
rm fs0:\startup.nsh

# Reset for the change to take effect
reset
EOF

	# Use serial port ttyAMA0
	cat > /etc/default/grub.d/99-qemu-efi-noacpi.cfg <<EOF
# Use serial port ttyAMAO
GRUB_CMDLINE_LINUX_DEFAULT="console=tty1 console=ttyAMA0"
EOF
fi

# Set the BootNext variable to the EFI Shell, causing it to be started on the
# next boot. On startup, the EFI Shell will automatically run startup.nsh
# (unless user-interrupted). Subsequent reboots will return to the normal
# BootOrder.
efibootmgr -n "${shell_bootnum}"

# Recreate the grub config file
update-grub

if [ -x /usr/share/update-notifier/notify-reboot-required ] ; then
	/usr/share/update-notifier/notify-reboot-required
fi
