#!/bin/sh
# Canonical/Ubuntu specific testbed setup

# Remove trailing dot from the machine fqdn.
# We also need to disable the update_etc_hosts cloud-init module:
# setting `manage_etc_hosts: false` in /e/c/cloud.cfg does not appear
# to be enough, as apparently the user-data can override that.
# Workaround for LP: #2019472.
sed -Ei '/^ +- update_etc_hosts/d' /etc/cloud/cloud.cfg
sed -Ei '/^127\.0\.1\.1[[:blank:]]/s/([[:alnum:]])\. /\1 /' /etc/hosts

RELEASE=$(lsb_release --codename --short 2>/dev/null)

# Make the installed package versions match what is available from the configured
# repositories, even if a downgrade is required.
#
# We may end up with images with packages newer than those available from the
# repositories when opening a new release (images are created using the
# previous release dailies as a base, which may include 0-day SRUs, or
# Launchpad may fail to copy packages, see openssl 3.0.8-1ubuntu2), or when
# package versions are pulled from the archive.
#
# Note: 'release a=' matches the Suite field in the repository Release file.
cat >/etc/apt/preferences.d/force-downgrade-to-release.pref <<EOF
Package: *
Pin: release a=$RELEASE
Pin-Priority: 1001

Package: *
Pin: release a=$RELEASE-updates
Pin-Priority: 1001

Package: *
Pin: release n=$RELEASE-security
Pin-Priority: 1001
EOF
# This is going to be a downgrade-only (or no-op) dist-upgrade, as setup-testbed
# already dist-upgraded the system. Force the usage of the conf files of the
# package being installed (--force-confnew), as conf files "from the future"
# may not be supported by the older package version we're downgrading to.
if [ "${RELEASE}" != "trusty" ]; then
    DEBIAN_FRONTEND=noninteractive eatmydata apt-get -y --allow-downgrades -o Dpkg::Options::="--force-confnew" dist-upgrade
fi
rm /etc/apt/preferences.d/force-downgrade-to-release.pref

# Remove packages not available from any configured repository, e.g.
# packages fully removed from the archive. This requires the analyze-pattern
# feature, which has been introduced in apt 2.0.1 (>= Focal).
#
# Note: autoremove/--autoremove always runs on the entire set of installed
# packages. Together with --autopurge/purge each removal becomes a purge.
#
# Removing '?obsolete' packages may remove the running kernel. This is not
# expected to be an issue.
if /usr/lib/apt/apt-helper analyze-pattern '?true' >/dev/null 2>&1; then
    DEBIAN_FRONTEND=noninteractive eatmydata apt-get -y purge --autoremove '?obsolete'
else
    # We're on an older release (analyze-pattern is not available).
    # Still do an autopurge to get rid of packages that got orphaned
    # by downgrades.
    eatmydata apt-get -y autoremove --purge
fi

# compatibility with old Jenkins testbeds for trusty
if grep -q trusty /etc/lsb-release; then apt-get update || (sleep 10; apt-get update); apt-get install -y build-essential; fi

# these packages really break tests, purge them if we use standard cloud images
# (on lcy01)
for p in unattended-upgrades lxd lxd-client snapd; do
    [ ! -d /usr/share/doc/$p ] || $(which eatmydata || true) apt-get purge --auto-remove -y $p
done

# work around broken PTMU; LP: #1572026, RT#90771
if type iptables >/dev/null 2>&1; then
    cat <<EOF > /etc/rc.local
#!/bin/sh
iptables -w -t mangle -A FORWARD -p tcp --tcp-flags SYN,RST SYN -j TCPMSS --clamp-mss-to-pmtu || true
EOF
    chmod 755 /etc/rc.local
    # shellcheck disable=SC1091
    . /etc/rc.local
fi

# work around broken tty on ppc64el instances (LP: #1607075)
if [ "$(uname -m)" = ppc64le ] && [ "$(systemctl is-active keyboard-setup.service 2>/dev/null|| true)" = failed ]; then
    systemctl reset-failed keyboard-setup.service
    systemctl mask keyboard-setup.service
fi

# work around late urandom pool initialization (LP: #1622893)
if [ -d /run/systemd/system ] && systemd-detect-virt --quiet --vm; then
    mkdir -p /etc/systemd/system/systemd-random-seed.service.d/
    cat <<EOF > /etc/systemd/system/systemd-random-seed.service.d/fakeentropy.conf
[Service]
ExecStart=/usr/bin/perl -E 'open \$\$f, "/bin/bash" or die; open \$\$rnd, ">/dev/random" or die; for (\$\$i = 0; \$\$i < 10; ++\$\$i) {read \$\$f, \$\$d, 64; ioctl \$\$rnd, 0x40085203, pack("ii", 64*8, 64) . \$\$d}'
EOF
fi

# set compressor to lz4 where supported, we care about compression speed
if grep -q lz4 /etc/initramfs-tools/initramfs.conf; then
    echo COMPRESS=lz4 > /etc/initramfs-tools/conf.d/lz4
fi
