# pylint: disable=missing-module-docstring,missing-function-docstring
import glob
import json
import os
import socket
import subprocess
from pathlib import Path
from textwrap import dedent

import pygit2
import yaml
from charmhelpers.core.hookenv import (
    charm_dir,
    config,
    log,
    storage_get,
    storage_list,
)
from charms.layer import status
from charms.reactive import (
    clear_flag,
    hook,
    not_unless,
    set_flag,
    when,
    when_all,
    when_any,
    when_not,
    when_not_all,
)
from utils import UnixUser, install_autodep8

AUTOPKGTEST_LOCATION = os.path.expanduser("~ubuntu/autopkgtest")
AUTOPKGTEST_CLOUD_LOCATION = os.path.expanduser("~ubuntu/autopkgtest-cloud")
AUTODEP8_LOCATION = os.path.expanduser("~ubuntu/autodep8")
AUTOPKGTEST_PER_PACKAGE_LOCATION = os.path.expanduser(
    "~ubuntu/autopkgtest-package-configs/"
)

AUTOPKGTEST_CLONE_LOCATION = (
    "https://salsa.debian.org/ubuntu-ci-team/autopkgtest.git"
)

AUTOPKGTEST_CLONE_BRANCH = "ubuntu/production"

AUTODEP8_CLONE_LOCATION = (
    "https://git.launchpad.net/~ubuntu-release/+git/autodep8"
)

AUTOPKGTEST_PER_PACKAGE_CLONE_LOCATION = (
    "https://git.launchpad.net/~ubuntu-release/autopkgtest"
    + "-cloud/+git/autopkgtest-package-configs"
)

RABBITMQ_CRED_PATH = os.path.expanduser("~ubuntu/rabbitmq.cred")


@hook("upgrade-charm")
def configure_vim():
    status.maintenance("Configuring vim")
    Path("/etc/vim/vimrc.local").write_text(
        """" autopkgtest-cloud vimrc, nicer for admins
" This file is written by the autopkgtest-cloud charm, do not edit manually, or
" you may loose your changes.
set background=dark                                             " lighter colors by default
set number                                                      " show line numbers
set listchars=eol:¬,tab:>·,trail:~,extends:>,precedes:<,nbsp:‡  " set display of special characters
set list                                                        " show those special characters
highlight ExtraWhitespace ctermbg=red guibg=red                 " highlight extra whitespaces
match ExtraWhitespace /\\s\\+$\\| \\+\\ze\\t/                         " define the extra whitespaces
    """
    )


@when_not("autopkgtest.autopkgtest_cloud_symlinked")
def symlink_autopkgtest_cloud():
    status.maintenance("Creating symlink to charmed autopkgtest-cloud code")
    with UnixUser("ubuntu"):
        try:
            autopkgtest_cloud = os.path.join(charm_dir(), "autopkgtest-cloud")
            os.symlink(autopkgtest_cloud, AUTOPKGTEST_CLOUD_LOCATION)
        except FileExistsError:
            pass
    status.maintenance(
        "Done creating symlink to charmed autopkgtest-cloud code"
    )
    set_flag("autopkgtest.autopkgtest_cloud_symlinked")


@when("apt.installed.git")
@when_not("autopkgtest.autodep8_cloned")
def clone_autodep8():
    status.maintenance("Cloning autodep8")
    log("Cloning autodep8", "INFO")
    with UnixUser("ubuntu"):
        try:
            pygit2.clone_repository(AUTODEP8_CLONE_LOCATION, AUTODEP8_LOCATION)
            log("autodep8 cloned", "INFO")
            status.maintenance("autodep8 cloned")
            set_flag("autopkgtest.autodep8_cloned")
        except ValueError:
            log("autodep8 already cloned", "INFO")


@when("apt.installed.git")
@when_not("autopkgtest.autopkgtest_package_configs_cloned")
def clone_autopkgtest_package_configs():
    status.maintenance("Cloning autopkgtest-package-configs")
    log("Cloning autopkgtest-package-configs", "INFO")
    with UnixUser("ubuntu"):
        try:
            pygit2.clone_repository(
                AUTOPKGTEST_PER_PACKAGE_CLONE_LOCATION,
                AUTOPKGTEST_PER_PACKAGE_LOCATION,
            )
            log("autopkgtest-package-configs cloned", "INFO")
            status.maintenance("autopkgtest-package-configs cloned")
            set_flag("autopkgtest.autopkgtest_package_configs_cloned")
        except ValueError:
            log("autopkgtest-package-configs already cloned", "INFO")


@when("autopkgtest.autodep8_cloned")
@when_not("autopkgtest.autodep8_installed")
def initially_install_autodep8():
    status.maintenance("Installing autodep8")
    install_autodep8(AUTODEP8_LOCATION)
    status.maintenance("Done installing autodep8")
    set_flag("autopkgtest.autodep8_installed")


@when("apt.installed.git")
@when_not("autopkgtest.autopkgtest_cloned")
def clone_autopkgtest():
    status.maintenance("Cloning autopkgtest")
    log("Cloning autopkgtest", "INFO")
    with UnixUser("ubuntu"):
        try:
            pygit2.clone_repository(
                AUTOPKGTEST_CLONE_LOCATION,
                AUTOPKGTEST_LOCATION,
                checkout_branch=AUTOPKGTEST_CLONE_BRANCH,
            )
            status.maintenance("autopkgtest cloned")
            log("autopkgtest cloned", "INFO")
            set_flag("autopkgtest.autopkgtest_cloned")
        except ValueError:
            # exists
            log("autopkgtest already cloned", "INFO")


@when_all(
    "autopkgtest.autopkgtest_cloud_symlinked",
    "autopkgtest.rabbitmq-configured",
    "autopkgtest.influx-creds-written",
)
def set_up_systemd_units():
    status.maintenance("Installing systemd units")
    for unit in glob.glob(os.path.join(charm_dir(), "units", "*")):
        base = os.path.basename(unit)
        dest = os.path.join(os.path.sep, "etc", "systemd", "system", base)

        def link_and_enable():
            # pylint: disable=cell-var-from-loop
            os.symlink(unit, dest)
            if "@" not in base:
                subprocess.check_call(["systemctl", "enable", base])

        try:
            link_and_enable()
        except FileExistsError:
            if not os.path.islink(dest):
                os.unlink(dest)
                link_and_enable()
    status.maintenance("Done installing systemd units")
    set_flag("autopkgtest.systemd_units_linked_and_enabled")


@when("autopkgtest.systemd_units_linked_and_enabled")
@when_not("autopkgtest.target_running")
def start():
    status.maintenance("Enabling and starting autopkgtest systemd target")
    subprocess.check_call(
        ["systemctl", "enable", "--now", "autopkgtest.target"]
    )
    status.maintenance("Done enabling and starting autopkgtest systemd target")
    set_flag("autopkgtest.target_running")


@hook("stop")
@when("autopkgtest.target_running")
def stop():
    status.maintenance("Disabling and stopping autopkgtest systemd target")
    subprocess.check_call(
        ["systemctl", "disable", "--now", "autopkgtest.target"]
    )
    status.maintenance(
        "Done disabling and stopping autopkgtest systemd target"
    )
    clear_flag("autopkgtest.target_running")


@when_all("autopkgtest.target-restart-needed", "autopkgtest.target_running")
def restart_target():
    status.maintenance("Restarting autopkgtest systemd target")
    subprocess.check_call(["systemctl", "restart", "autopkgtest.target"])
    status.maintenance("Done restarting autopkgtest systemd target")
    clear_flag("autopkgtest.target-restart-needed")


@when_not("autopkgtest.target-restart-needed", "apt.queued_installs")
@when("autopkgtest.target_running")
def is_active():
    status.active("autopkgtest workers running and processing requests")


@when_all(
    "autopkgtest.systemd_units_linked_and_enabled",
    "autopkgtest.daemon-reload-needed",
)
@hook("upgrade-charm")
def daemon_reload():
    status.maintenance("Reloading systemd configuration after change")
    subprocess.check_call(["systemctl", "daemon-reload"])
    status.maintenance("Done reloading systemd configuration after change")
    clear_flag("autopkgtest.daemon-reload-needed")


@hook("upgrade-charm")
def install_apt_packages():
    """Install packages defined in layer.yaml"""
    set_flag("apt.needs_update")
    clear_flag("apt.queued_installs")


@hook("install")
def enable_persistent_journal():
    status.maintenance("Enabling persistent journal")
    try:
        journal_dir = os.path.join(os.path.sep, "var", "log", "journal")
        os.makedirs(journal_dir)
        subprocess.check_call(
            ["systemd-tmpfiles", "--create", "--prefix", journal_dir]
        )
        subprocess.check_call(["systemctl", "restart", "systemd-journald"])
    except FileExistsError:
        pass
    status.maintenance("Done enabling persistent journal")


@when("amqp.connected")
@when_not("amqp.available")
def setup_rabbitmq(rabbitmq):
    rabbitmq.request_access("autopkgtest-worker", "/")
    status.waiting("Waiting on RabbitMQ to configure vhost")


@when("amqp.available")
@when_not("autopkgtest.rabbitmq-configured")
def set_up_rabbitmq(rabbitmq):
    username = rabbitmq.username()
    password = rabbitmq.password()
    host = rabbitmq.private_address()
    status.maintenance("Configuring rabbitmq")
    log("Setting up rabbitmq connection to: {}@{}".format(username, host))
    with open(RABBITMQ_CRED_PATH, "w") as cred_file:
        cred_file.write(
            dedent(
                """\
            RABBIT_HOST="{host}"
            RABBIT_USER="{username}"
            RABBIT_PASSWORD="{password}"
                """.format(
                    **locals()
                )
            )
        )
    status.maintenance("rabbitmq configured")
    set_flag("autopkgtest.rabbitmq-configured")


@when_not("amqp.available")
def clear_rabbitmq():
    status.maintenance("clearing rabbitmq configuration")
    try:
        log("rabbitmq not available, deleting credentials file")
        os.unlink(RABBITMQ_CRED_PATH)
        clear_flag("autopkgtest.rabbitmq-configured")
    except FileNotFoundError:
        pass
    status.maintenance("Done clearing rabbitmq configuration")


@when("config.changed.nova-rcs")
def update_nova_rcs():
    status.maintenance("Updating nova rc files")
    # pylint: disable=import-outside-toplevel
    import base64
    from io import BytesIO
    from tarfile import TarFile

    rctar = config().get("nova-rcs")

    if not rctar:
        # set to the empty value
        return

    log("Extracting cloud .rc files...", "INFO")

    clear_old_rcs()

    bytes_file = BytesIO(base64.b64decode(rctar))
    tar = TarFile(fileobj=bytes_file)

    log("...got {}".format(", ".join(tar.getnames())), "INFO")

    tar.extractall(os.path.expanduser("~ubuntu/cloudrcs/"))
    status.maintenance("Done updating nova rc files")


@when("config.default.nova-rcs")
def clear_old_rcs():
    status.maintenance("Cleaning old nova rc files")
    rcfiles = glob.glob(os.path.expanduser("~ubuntu/cloudrcs/*.rc"))

    if not rcfiles:
        return

    log("Deleting old cloud .rc files...", "INFO")

    for rcfile in rcfiles:
        log("...deleting {}...".format(rcfile))
        os.unlink(rcfile)

    log("...done", "INFO")
    status.maintenance("Done cleaning old nova rc files")


@when_all(
    "autopkgtest.autopkgtest_cloud_symlinked",
    "autopkgtest.systemd_units_linked_and_enabled",
    "config.set.releases",
)
@when_any(
    "config.set.n-workers",
    "config.set.lxd-remotes",
)
@when_not("autopkgtest.systemd-units-initially-enabled")
def enable_units_initially():
    enable_disable_units()


@when_any(
    "config.set.lxd-remotes",
)
def remove_old_lxd_units():
    # pylint: disable=import-outside-toplevel
    from lib.systemd import remove_old_lxd_dropins

    lxdremotes = config().get("lxd-remotes") or ""
    lxdremotes_yaml = yaml.load(lxdremotes, Loader=yaml.CSafeLoader)
    log("Removing old lxd systemd units for old remotes")
    remove_old_lxd_dropins(lxdremotes_yaml)
    set_flag("autopkgtest.reload-needed")
    set_flag("autopkgtest.daemon-reload-needed")
    set_flag("autopkgtest.target-restart-needed")


@when_all(
    "autopkgtest.autopkgtest_cloud_symlinked",
    "autopkgtest.systemd_units_linked_and_enabled",
    "config.set.releases",
)
@when_any(
    "config.changed.n-workers",
    "config.set.lxd-remotes",
    "config.changed.lxd-remotes",
    "config.changed.releases",
)
def enable_disable_units():
    # pylint: disable=import-outside-toplevel
    from lib.systemd import set_up_systemd_units

    nworkers = config().get("n-workers") or ""
    lxdremotes = config().get("lxd-remotes") or ""
    releases = config().get("releases") or ""

    nworkers_yaml = yaml.load(nworkers, Loader=yaml.CSafeLoader)
    lxdremotes_yaml = yaml.load(lxdremotes, Loader=yaml.CSafeLoader)

    log("Setting up systemd worker units", "INFO")
    status.maintenance("Enabling and starting worker units")
    set_up_systemd_units(
        nworkers_yaml or {}, lxdremotes_yaml or {}, releases.split()
    )
    if lxdremotes_yaml != {}:
        with open("/home/ubuntu/lxd-remotes.json", "w") as f:
            json.dump(lxdremotes_yaml, f)

    status.maintenance("Done enabling and starting worker units")

    set_flag("autopkgtest.reload-needed")
    set_flag("autopkgtest.daemon-reload-needed")
    set_flag("autopkgtest.target-restart-needed")
    set_flag("autopkgtest.systemd-units-initially-enabled")


@when("apt.installed.lxd-client")
@when_any(
    "config.set.n-workers", "config.set.lxd-remotes", "config.set.releases"
)
@when_not("autopkgtest.ubuntu_added_to_lxd_group")
def add_ubuntu_user_to_lxd_group():
    status.maintenance("Adding ubuntu user to lxd group")
    subprocess.check_call(["adduser", "ubuntu", "lxd"])
    status.maintenance("Done adding ubuntu user to lxd group")
    set_flag("autopkgtest.ubuntu_added_to_lxd_group")


@when("config.changed.swift-auth-version")
def auth_version_changed():
    auth_version = config().get("swift-auth-version")

    if auth_version not in [2, 3]:
        raise ValueError("Only auth versions 2 and 3 are supported")

    if auth_version == 2:
        set_flag("autopkgtest.v2_auth")
        clear_flag("autopkgtest.v3_auth")
    else:
        clear_flag("autopkgtest.v2_auth")
        set_flag("autopkgtest.v3_auth")


@not_unless("autopkgtest.v3_auth")
@when_any(
    "config.changed.swift-auth-url",
    "config.changed.swift-username",
    "config.changed.swift-password",
    "config.changed.swift-region",
    "config.changed.swift-project-domain-name",
    "config.changed.swift-project-name",
    "config.changed.swift-user-domain-name",
    "config.changed.swift-auth-version",
)
@when_all(
    "config.set.swift-auth-url",
    "config.set.swift-username",
    "config.set.swift-password",
    "config.set.swift-region",
    "config.set.swift-project-domain-name",
    "config.set.swift-project-name",
    "config.set.swift-user-domain-name",
    "config.set.swift-auth-version",
)
def write_v3_config():
    write_swift_config()


@not_unless("autopkgtest.v2_auth")
@when_any(
    "config.changed.swift-auth-url",
    "config.changed.swift-username",
    "config.changed.swift-password",
    "config.changed.swift-region",
    "config.changed.swift-auth-version",
    "config.changed.swift-tenant",
)
@when_all(
    "config.set.swift-auth-url",
    "config.set.swift-username",
    "config.set.swift-password",
    "config.set.swift-region",
    "config.set.swift-auth-version",
    "config.set.swift-tenant",
)
def write_v2_config():
    write_swift_config()


def write_swift_config():
    status.maintenance("Writing swift configuration")
    with open(
        os.path.expanduser("~ubuntu/swift-password.cred"), "w"
    ) as swift_password_file:
        for key in config():
            if key.startswith("swift") and config()[key] is not None:
                swift_password_file.write(
                    '{}="{}"\n'.format(
                        key.upper().replace("-", "_"),
                        str(config()[key]).strip(),
                    )
                )
    status.maintenance("Done writing swift configuration")


@when_any(
    "config.changed.worker-default-flavor",
    "config.changed.worker-big-flavor",
    "config.changed.worker-args",
    "config.changed.worker-setup-command",
    "config.changed.worker-setup-command2",
    "config.changed.releases",
    "config.changed.n-workers",
    "config.changed.lxd-remotes",
    "config.changed.mirror",
    "config.changed.net-name",
    "config.changed.worker-upstream-percentage",
)
@when_any("config.set.nova-rcs", "config.set.lxd-remotes")
def write_worker_config():
    status.maintenance("Writing worker configuration")

    # pylint: disable=import-outside-toplevel
    import configparser

    nworkers = config().get("n-workers") or ""
    lxdremotes = config().get("lxd-remotes") or ""

    if not nworkers and not lxdremotes:
        return

    nworkers_yaml = yaml.load(nworkers, Loader=yaml.CSafeLoader) or {}
    lxdremotes_yaml = yaml.load(lxdremotes, Loader=yaml.CSafeLoader) or {}

    conf = {
        "autopkgtest": {
            "checkout_dir": AUTOPKGTEST_LOCATION,
            "releases": config().get("releases"),
            "setup_command": config().get("worker-setup-command"),
            "setup_command2": config().get("worker-setup-command2"),
            "per_package_config_dir": AUTOPKGTEST_PER_PACKAGE_LOCATION,
            "worker_upstream_percentage": config().get(
                "worker-upstream-percentage"
            ),
        },
        "virt": {
            "package_size_default": config().get("worker-default-flavor"),
            "package_size_big": config().get("worker-big-flavor"),
            "args": config().get("worker-args"),
        },
    }

    def subst(s):
        replacements = {
            "/CHECKOUTDIR/": AUTOPKGTEST_LOCATION,
            "/HOSTNAME/": socket.gethostname(),
            "/AUTOPKGTEST_CLOUD_DIR/": AUTOPKGTEST_CLOUD_LOCATION,
            "/MIRROR/": config().get("mirror"),
            "/NET_NAME/": config().get("net-name"),
        }

        for k in replacements:  # pylint: disable=consider-using-dict-items
            if replacements[k]:
                if isinstance(s, str):
                    s = s.replace(k, replacements[k])

        return s

    # prune anything which wasn't set (nested dict)
    conf = {
        k: {l: subst(w) for l, w in v.items() if w is not None}
        for k, v in conf.items()
    }

    def write(conf_file):
        with open(conf_file, "w") as cf:
            cp = configparser.ConfigParser()
            cp.read_dict(conf)
            cp.write(cf)

        # FIXME: Hotfix for bos01
        if "bos01" in conf_file:
            with open(conf_file, "r") as cf:
                conf_data = cf.read()
            with open(conf_file, "w") as cf:
                cf.write(
                    conf_data.replace(
                        config().get("mirror"),
                        "http://us.ports.ubuntu.com/ubuntu-ports/",
                    )
                )

    for region in nworkers_yaml:
        for arch in nworkers_yaml[region]:
            if arch == "amd64":
                conf_file = os.path.join(
                    os.path.expanduser("~ubuntu"),
                    "worker-{}.conf".format(region),
                )
                conf["autopkgtest"]["architectures"] = "amd64 i386"
                write(conf_file)
                break
            else:
                conf_file = os.path.join(
                    os.path.expanduser("~ubuntu"),
                    "worker-{}-{}.conf".format(region, arch),
                )
                conf["autopkgtest"]["architectures"] = arch
                write(conf_file)

    for arch in lxdremotes_yaml:
        conf_file = os.path.join(
            os.path.expanduser("~ubuntu"), "worker-lxd-{}.conf".format(arch)
        )
        conf["autopkgtest"]["architectures"] = arch
        write(conf_file)

    status.maintenance("Done writing worker configuration")
    set_flag("autopkgtest.daemon-reload-needed")
    set_flag("autopkgtest.reload-needed")


@when("config.changed.net-name")
def write_net_name():
    status.maintenance("Writing net-name configuration")
    clear_flag("autopkgtest.net-name-written")
    with open(os.path.expanduser("~ubuntu/net-name.rc"), "w") as f:
        f.write('NET_NAME="{}"\n'.format(config().get("net-name")))
    status.maintenance("Done writing net-name configuration")
    set_flag("autopkgtest.net-name-written")
    set_flag("autopkgtest.reload-needed")


@when("config.changed.mirror")
def write_mirror():
    status.maintenance("Writing mirror configuration")
    with open(os.path.expanduser("~ubuntu/mirror.rc"), "w") as f:
        f.write('MIRROR="{}"\n'.format(config().get("mirror")))
    status.maintenance("Done writing mirror configuration")
    set_flag("autopkgtest.reload-needed")


@when("autopkgtest.reload-needed")
@when_not("autopkgtest.daemon-reload-needed")
def reload_systemd_units():
    # pylint: disable=import-outside-toplevel
    from lib.systemd import reload_units

    reload_units()
    clear_flag("autopkgtest.reload-needed")


@hook("tmp-storage-attached")
def fix_tmp_permissions():
    status.maintenance("Fixing tmp permissions")
    storageids = storage_list("tmp")
    if not storageids:
        status.blocked("Cannot locate attached storage")
        return
    storageid = storageids[0]

    mount = storage_get("location", storageid)

    os.chmod(mount, 0o777)
    status.maintenance("Done fixing tmp permissions")


@when_any(
    "config.set.worker-tmp-cleanup-config",
    "config.changed.worker-tmp-cleanup-config",
)
def write_worker_tmp_cleanup_config():
    cleanup_cfg_str = config().get("worker-tmp-cleanup-config")
    with open("/etc/tmpfiles.d/tmp.conf", "w") as f:
        f.write("%s\n" % cleanup_cfg_str)


@when_any(
    "config.changed.influxdb-hostname",
    "config.changed.influxdb-port",
    "config.changed.influxdb-username",
    "config.changed.influxdb-password",
    "config.changed.influxdb-database",
    "config.changed.influxdb-context",
)
@when_all(
    "config.set.influxdb-hostname",
    "config.set.influxdb-port",
    "config.set.influxdb-username",
    "config.set.influxdb-password",
    "config.set.influxdb-database",
    "config.set.influxdb-context",
)
def write_influx_creds():
    status.maintenance("Writing influxdb credentials")
    influxdb_hostname = config().get("influxdb-hostname")
    influxdb_port = config().get("influxdb-port")
    influxdb_username = config().get("influxdb-username")
    influxdb_password = config().get("influxdb-password")
    influxdb_database = config().get("influxdb-database")
    influxdb_context = config().get("influxdb-context")

    with open(os.path.expanduser("~ubuntu/influx.cred"), "w") as cf:
        cf.write(
            dedent(
                f"""\
        INFLUXDB_HOSTNAME={influxdb_hostname}
        INFLUXDB_PORT={influxdb_port}
        INFLUXDB_USERNAME={influxdb_username}
        INFLUXDB_PASSWORD={influxdb_password}
        INFLUXDB_DATABASE={influxdb_database}
        INFLUXDB_CONTEXT={influxdb_context}
        """
            )
        )
    status.maintenance("Done writing influxdb credentials")
    set_flag("autopkgtest.influx-creds-written")


@when_not_all(
    "config.set.influxdb-hostname",
    "config.set.influxdb-port",
    "config.set.influxdb-username",
    "config.set.influxdb-password",
    "config.set.influxdb-database",
    "config.set.influxdb-context",
)
def unset_influx_creds():
    status.maintenance("Deleting influxdb credentials")
    try:
        os.unlink(os.path.expanduser("~ubuntu/influx.cred"))
    except FileNotFoundError:
        pass
    status.maintenance("Done deleting influxdb credentials")
    clear_flag("autopkgtest.influx-creds-written")
